/*
 * Initial main.c for Xdialog.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <unistd.h>   
#include <getopt.h>  
#include <stdio.h>
#include <string.h>
#include <stdlib.h>   

#include "interface.h"
#include "support.h"
#include "callbacks.h"

/* A structure used to pass Xdialog parameters. */
Xdialog_data Xdialog;

/* Usage displaying */

#define HELP_TEXT1 \
"Xdialog v"VERSION" by Thierry Godefroy <xdialog@free.fr> (v1.0 was\n\
written by Alfred at Cyberone Internet <alfred@cyberone.com.au>).\n\
Xdialog home page available at: http://xdialog.free.fr.\n\
\n\
Usage: "

#define HELP_TEXT2 \
" [<common options>] [<transient options>] <box option> ...\n\
\n\
Common options:\n\
  --wmclass <name>\n\
  --rc-file <gtkrc filename>\n\
  --backtitle <backtitle>\n\
  --title <title>\n\
  --allow-close | --no-close\n\
  --screen-center | --under-mouse | --auto-placement\n\
  --center | --right | --left\n\
  --no-wrap | --wrap\n\
  --stderr | --stdout\n\
  --separator <character> | --separate-output\n\
  --buttons-style default|icon|text\n\
\n\
Transient options:\n\
  --fixed-font\n\
  --password (may be repeated 2 or 3 times before --2inputsbox or --3inputsbox)\n\
  --interval <timeout>\n\
  --no-tags\n\
  --icon <xpm filename>\n\
  --no-cancel\n\
  --no-buttons\n\
  --default-no\n\
  --wizard\n\
  --help <help>\n\
  --print <printer> (1)\n\
\n\
Box options:\n\
  --yesno	<text> <height> <width>\n\
  --msgbox	<text> <height> <width>\n\
  --infobox	<text> <height> <width> [<timeout>]\n\
  --gauge	<text> <height> <width> [<percent>]\n\
  --inputbox	<text> <height> <width> [<init>]\n\
  --2inputsbox	<text> <height> <width> <label1> <init1> <label2> <init2>\n\
  --3inputsbox	<text> <height> <width> <label1> <init1> <label2> <init2> <label3> <init3>\n\
  --rangebox	<text> <height> <width> <min value> <max value> [<default value>]\n\
  --2rangesbox	<text> <height> <width> <label1> <min1> <max1> <def1> <label2> <min2> <max2> <def2>\n\
  --3rangesbox	<text> <height> <width> <label1> <min1> <max1> <def1> ... <label3> <min3> <max3> <def3>\n\
  --textbox	<file> <height> <width>\n\
  --editbox	<file> <height> <width>\n\
  --tailbox	<file> <height> <width>\n\
  --menubox	<text> <height> <width> <menu height> <tag1> <item1>...\n\
  --checklist	<text> <height> <width> <list height> <tag1> <item1> <status1>...\n\
  --radiolist	<text> <height> <width> <list height> <tag1> <item1> <status1>...\n\
  --buildlist	<text> <height> <width> <list height> <tag1> <item1> <status1>...\n\
  --treeview	<text> <height> <width> <list height> <tag1> <item1> <status1> <item_depth1>...\n\
  --fselect	<file> <height> <width>\n\
  --dselect	<directory> <height> <width>\n\
  --calendar	<text> <height> <width> <day> <month> <year>\n\
  --timebox	<text> <height> <width>\n\
\n\
Special options:\n\
  --version		(prints version number to stderr and exits).\n\
  --print-version	(same as above in a cdialog-compatible way).\n\
\n\
Note that <height> and <width> are in characters and may be replaced by a single\n\
<XSIZExYSIZE> parameter (like the one passed in the -geometry option of X) which\n\
will represent the size of the Xdialog window in pixels. Specifying a size of 0 0\n\
(or 0x0) will auto-size Xdialog.\n\
\n\
(1) This Xdialog version compiled with: "PRINTER_CMD" "PRINTER_CMD_OPTION"<printer> "TEMP_FILE"\n\
    as the print command. If <printer> is \"\" (an empty string), the "PRINTER_CMD_OPTION"\n\
    option is not used.\n\n"

#define HELP_MSG_SIZE 2900

static void print_help_info(char *name, char *errmsg)
{
	gchar msg[HELP_MSG_SIZE];
	gchar cmd[32];

	strcpy(cmd, strlen(name) < 32 ? name : XDIALOG);

	strcpysafe(msg, HELP_TEXT1, HELP_MSG_SIZE);
	strcatsafe(msg, cmd, HELP_MSG_SIZE);
	strcatsafe(msg, HELP_TEXT2, HELP_MSG_SIZE);

	fprintf(stderr, "%s: %s !\n", cmd, errmsg);
	fprintf(stderr, msg);

	if (strlen(msg) == HELP_MSG_SIZE-1)
		fprintf(stderr, "\n\nHelp message truncated, please re-compile "\
				"after increasing HELP_MSG_SIZE in main.c !\n");

	strcpysafe(Xdialog.title, "Usage for ", MAX_TITLE_LENGTH);
	strcatsafe(Xdialog.title, cmd, MAX_TITLE_LENGTH);
	Xdialog.cancel_button = Xdialog.help = Xdialog.icon = FALSE;
	if (!Xdialog.print) {
		Xdialog.print = TRUE;
		Xdialog.printer[0] = 0;
	}
	Xdialog.fixed_font = Xdialog.buttons = TRUE;
	Xdialog.backtitle[0] = 0;
	Xdialog.size_in_pixels = FALSE;
	Xdialog.xsize = 110;
	Xdialog.ysize = 42;
	create_textbox("", FALSE);
	gtk_text_insert(GTK_TEXT(Xdialog.widget1), NULL, NULL, NULL, msg,
			strlen(msg));
	gtk_widget_show(Xdialog.window);
	gtk_main();

	exit(255);
}

/* Xsize and Ysize parameters */

static void get_box_size(int argc, char *argv[], int *optind)
{
	char *index;

	if (*optind > argc-1)
		print_help_info(argv[0], "box size missing");

	if ((index = strstr(argv[*optind], "x")) == NULL &&
	    argv[*optind + 1] == NULL)
		print_help_info(argv[0], "bad size parameter");

	if (index == NULL) {
		Xdialog.size_in_pixels = FALSE;
		Xdialog.ysize = atoi(argv[(*optind)++]);
		Xdialog.xsize = atoi(argv[(*optind)++]);
	} else {
		Xdialog.size_in_pixels = TRUE;
		Xdialog.xsize = atoi(argv[(*optind)++]);
		Xdialog.ysize = atoi(++index);
	}

	if (Xdialog.xsize < 0 || Xdialog.ysize < 0)
		print_help_info(argv[0], "negative box size");
}

static int get_list_size(int argc, char *argv[], int *optind, int parameters,
			 gboolean with_list_height)
{
	int list_size, next_opt, i;

	i = *optind;
	if (with_list_height)
		i++;

	if (i >= argc)
		print_help_info(argv[0], "missing list parameters");

	if (with_list_height) {
		Xdialog.list_height = atoi(argv[(*optind)++]);
		if (Xdialog.list_height < 0)
			print_help_info(argv[0], "negative list height");
	}

	next_opt = argc;

	for (i = *optind; i < argc-1; i++)
		if (strstr(argv[i], "--") == argv[i]) {
			next_opt = i;
			break;
		}

	list_size = (next_opt - *optind) / parameters;

	if (list_size * parameters + *optind != next_opt)
		print_help_info(argv[0],
				"incorrect number of parameters in list");

	return list_size;
}

/* Main routine */

int main(int argc, char *argv[])
{
	gchar title_tmp[MAX_TITLE_LENGTH];
	gchar backt_tmp[MAX_BACKTITLE_LENGTH];
	gchar help_text[MAX_LABEL_LENGTH];
	gint option_index = 0;
	gint old_optind = 0;
	gint optcode;
	gint win = FALSE;
	gint timeout;
	gint percent;
	gint list_size = 0;
	gint min, max, deflt, day, month, year, i;
	gboolean icon_tmp;
	static struct option long_options[] = {
					/* Box options */	
						{"yesno", 1, 0, 0},
						{"msgbox", 1, 0, 1},
						{"infobox", 1, 0, 2},
						{"inputbox", 1, 0, 3},
						{"2inputsbox", 1, 0, 4},
						{"3inputsbox", 1, 0, 5},
						{"textbox", 1, 0, 6},
						{"editbox", 1, 0, 7},
						{"menubox", 1, 0, 8},
						{"checklist", 1, 0, 9},
						{"radiolist", 1, 0, 10},
						{"buildlist",1,0,11},
						{"treeview", 1, 0, 12},
						{"gauge",1,0,13},
						{"guage",1,0,13},
						{"tailboxbg",1,0,14},
						{"fselect",1,0,15},
						{"dselect",1,0,16},
						{"rangebox",1,0,17},
						{"2rangesbox",1,0,18},
						{"3rangesbox",1,0,19},
						{"calendar",1,0,20},
						{"timebox",1,0,21},
					/* Common options */	
						{"title",1,0,40},
						{"backtitle",1,0,41},
						{"wmclass",1,0,42},
						{"screen-center",0,0,43},
						{"under-mouse",0,0,44},
						{"auto-placement",0,0,45},
						{"left",0,0,46},
						{"right",0,0,47},
						{"center",0,0,48},
					        {"wrap",0,0,49},
						{"no-wrap",0,0,50},
						{"stdout",0,0,51},
						{"stderr",0,0,52},
						{"no-close",0,0,53},
						{"allow-close",0,0,54},
						{"buttons-style",1,0,55},
						{"rc-file",1,0,56},
						{"separator",1,0,57},
						{"separate-output",0,0,58},
					/* Transient options */	
						{"fixed-font",0,0,80},
						{"password",0,0,81},
						{"no-cancel",0,0,82},
						{"no-buttons",0,0,83},
						{"no-tags",0,0,84},
						{"help",1,0,85},
						{"print",1,0,86},
						{"wizard",0,0,87},
						{"default-no",0,0,88},
						{"defaultno",0,0,88},
						{"icon",1,0,89},
						{"interval",1,0,90},
					/* Special options */	
						{"version",0,0,98},
						{"print-version",0,0,99},
					/* End of options marker */	
						{0, 0, 0, 0}
					      };
	min = max = deflt = day = month = year = 0;
	opterr		= 0;

	memset(&Xdialog, 0, sizeof(Xdialog_data));

	Xdialog.output		= stderr;		/* Default output for Xdialog results */
	Xdialog.placement	= WINDOW_CENTERED;	/* Center window on screen as default */
	Xdialog.justify		= JUSTIFY_CENTER;	/* Center messages as default */
	Xdialog.wrap		= FALSE;		/* Don't wrap back-title and text as default */
	Xdialog.fixed_font	= FALSE;		/* Don't use a fixed font as default */
	Xdialog.passwd		= 0;			/* Don't use passwd input as default */
	Xdialog.interval	= 0;			/* Don't report periodically as default */
	Xdialog.buttons_style	= ICON_AND_TEXT;	/* Default buttons style (icon+text) */
	Xdialog.icon		= FALSE;		/* No icon as default */
	Xdialog.no_close	= FALSE;		/* Allow to close the box as default */
	Xdialog.buttons		= TRUE;			/* Display buttons as default */
	Xdialog.cancel_button	= TRUE;			/* Display "Cancel" button as default */
	Xdialog.help		= FALSE;		/* No help button as default */
	Xdialog.wizard		= FALSE;		/* No wizard buttons as default */
	Xdialog.default_no	= FALSE;		/* Default selected button is Yes/OK */
	Xdialog.print		= FALSE;		/* No print button as default */
	Xdialog.tags		= TRUE;			/* Display tags before items in lists */
	Xdialog.backtitle[0]	= 0;			/* Defaults to no backtitle */
	strcpy(Xdialog.title, XDIALOG);			/* Default widget title */
	Xdialog.rc_file[0]	= 0;			/* Defaults to no rc-file */
	strcpy(Xdialog.separator, "/");			/* Default results separator */

	if (gtk_major_version < 1 ||
	    (gtk_major_version == 1 && gtk_minor_version < 2))
		fprintf(stderr,
			"%s: GTK+ version too old, please upgrade !\n", argv[0]);
	gtk_set_locale();
	if (!gtk_init_check(&argc, &argv)) {
		fprintf(stderr,
			"%s: Error initializing the GUI...\n"
			"Do you run under X11 with GTK+ v1.2.0+ installed ?\n", argv[0]);
		return 255;
	}

#if 0	/* deprecated */
	add_pixmap_directory(PACKAGE_DATA_DIR "/pixmaps");
	add_pixmap_directory(PACKAGE_SOURCE_DIR "/pixmaps");
#endif

	while (TRUE) {
		optcode = getopt_long_only(argc, argv, "", long_options, &option_index);

		if (optcode == -1)
			break;
		/* if it is a box option or the "--backtitle" / "--help" / "--separator"
		 * options, then translates the "\n" into newlines...
		 */
		if (optcode > -1 && (optcode < 40 || optcode == 41
				     || optcode == 57 || optcode == 85)) {
			char *tmp;

			while (strstr(optarg, "\\n") != NULL) {
				tmp = strstr(optarg, "\\n");
				strcpy(tmp, tmp+1);
				*tmp = '\n';
			}
		}

		old_optind = optind;
show_again:
		win = FALSE;

		switch (optcode) {
		/* Box options */
			case 0: /* a yesno box */
				get_box_size(argc, argv, &optind);
				create_msgbox(optarg, TRUE);
				win = TRUE;
				break;
			case 1: /* a msgbox */
				get_box_size(argc, argv, &optind);
				create_msgbox(optarg, FALSE);
				win = TRUE;
				break;
			case 2: /* an infobox */
				get_box_size(argc, argv, &optind);
				if (argv[optind] != NULL)
					timeout = atoi(argv[optind++]);
				else
					timeout	= INFO_TIME;
				create_infobox(optarg, timeout);
				win = TRUE;
				break;
			case 3: /* an inputbox */
			case 4: /* a 2inputsbox */
			case 5: /* a 3inputsbox */
				get_box_size(argc, argv, &optind);
				if (optcode != 3) {
					list_size = get_list_size(argc, argv, &optind, 2, FALSE);
					if (list_size != optcode - 2)
						print_help_info(argv[0],
								"incorrect number of parameters");
				}
				create_inputbox(optarg, argv+optind, optcode - 2);
				if (optcode == 3)
					optind += ((argv[optind] != NULL) ? 1 : 0);
				else
					optind += 2*list_size;
				win = TRUE;
				break;
			case 6: /* a text box */
				get_box_size(argc, argv, &optind);
				create_textbox(optarg, FALSE);
				win = TRUE;
				break;
			case 7: /* an edit box */
				get_box_size(argc, argv, &optind);
				create_textbox(optarg, TRUE);
				win = TRUE;
				break;
			case 8: /* a menu box */
				get_box_size(argc, argv, &optind);
				list_size = get_list_size(argc, argv, &optind, 2, TRUE);
				create_menubox(optarg, argv+optind, list_size);
				optind += 2*list_size;
				win = TRUE;
				break;
			case 9: /* a check list */
				get_box_size(argc, argv, &optind);
				list_size = get_list_size(argc, argv, &optind, 3, TRUE);
				create_itemlist(optarg, CHECKLIST, argv+optind,
						list_size);
				optind += 3*list_size;
				win = TRUE;
				break;
			case 10: /* a radio list */
				get_box_size(argc, argv, &optind);
				list_size = get_list_size(argc, argv, &optind, 3, TRUE);
				create_itemlist(optarg, RADIOLIST, argv+optind,
						list_size);
				optind += 3*list_size;
				win = TRUE;
				break;
			case 11: /* a build list */
				get_box_size(argc, argv, &optind);
				list_size = get_list_size(argc, argv, &optind, 3, TRUE);
				create_buildlist(optarg, argv+optind, list_size);
				optind += 3*list_size;
				win = TRUE;
				break;
			case 12: /* a tree view */
				get_box_size(argc, argv, &optind);
				list_size = get_list_size(argc, argv, &optind, 4, TRUE);
				create_treeview(optarg, argv+optind, list_size);
				optind += 4*list_size;
				win = TRUE;
				break;
			case 13: /* a gauge (or "guage", typo in c/gdialog kept for compatibility... */
				get_box_size(argc, argv, &optind);
				if (argv[optind] == NULL)
					percent = 0;
				else
					percent = atoi(argv[optind++]);
				create_gauge(optarg, percent);
				win = TRUE;
				break;
			case 14: /* a tail boxes */
				get_box_size(argc, argv, &optind);
				create_tailbox(optarg);
				win = TRUE;
				break;
			case 15: /* file selector */
				get_box_size(argc, argv, &optind);
				create_filesel(optarg, FALSE);
				win = TRUE;
				break;
			case 16: /* directory selector */
				get_box_size(argc, argv, &optind);
				create_filesel(optarg, TRUE);
				win = TRUE;
				break;
			case 17: /* a range box */
			case 18: /* a 2ranges box */
			case 19: /* a 3ranges box */
				get_box_size(argc, argv, &optind);
				if (optcode != 17) {
					list_size = get_list_size(argc, argv, &optind, 4, FALSE);
					if (list_size != optcode - 16)
						print_help_info(argv[0],
								"incorrect number of parameters");
					for (i = 0 ; i < list_size; i++) {
						min   = atoi(argv[optind+4*i+1]);
						max   = atoi(argv[optind+4*i+2]);
						deflt = atoi(argv[optind+4*i+3]);
						if (min >= max || deflt < min || deflt > max)
							print_help_info(argv[0],
									"bad range parameters");
					}
				} else {
					if (argv[optind] == NULL || argv[optind+1] == NULL)
						print_help_info(argv[0],
								"missing range parameter");
					else {
						min = atoi(argv[optind]);
						max = atoi(argv[optind+1]);
					}
					if (argv[optind+2] == NULL)
						deflt = min;
					else
						deflt = atoi(argv[optind+2]);
					if (min >= max || deflt < min || deflt > max)
						print_help_info(argv[0],
								"bad range parameters");
				}
				create_rangebox(optarg, argv+optind, optcode - 16);
				if (optcode != 17)
					optind += 4*list_size;
				else
					optind += 2 + (argv[optind+2] != NULL ? 1 : 0);
				win = TRUE;
				break;
			case 20: /* a calendar */
				get_box_size(argc, argv, &optind);
				if (argv[optind] == NULL ||
				    argv[optind+1] == NULL ||
				    argv[optind+2] == NULL)
					print_help_info(argv[0],
							"missing calendar parameter");
				else {
					day   = atoi(argv[optind++]);
					month = atoi(argv[optind++]);
					year  = atoi(argv[optind++]);
				}
				if (day < 0 || day > 31 ||
				    month < 0 || month > 12 ||
				    (year != 0 && year < 1970))
					print_help_info(argv[0],
							"bad calendar date");
				create_calendar(optarg, day, month, year);
				win = TRUE;
				break;
			case 21: /* a time box */
				get_box_size(argc, argv, &optind);
				create_timebox(optarg);
				win = TRUE;
				break;
		/* Common options */
			case 40: /* --title option */
				strcpysafe(Xdialog.title, optarg,
					   MAX_TITLE_LENGTH);
	  			break;
			case 41: /* --backtitle option */
				strcpysafe(Xdialog.backtitle, optarg,
					   MAX_BACKTITLE_LENGTH);
				break;
			case 42: /* --wmclass option */
				strcpysafe(Xdialog.wmclass, optarg,
					   MAX_WMCLASS_LENGTH);
				break;
			case 43: /* --screen-center option */
				Xdialog.placement = WINDOW_CENTERED;
				break;
			case 44: /* --under-mouse option */
				Xdialog.placement = WINDOW_UNDER_MOUSE;
				break;
			case 45: /* --auto-placement option */
				Xdialog.placement = WINDOW_AUTO_PLACEMENT;
				break;
			case 46: /* --left option */
				Xdialog.justify = JUSTIFY_LEFT;
				break;
			case 47: /* --right option */
				Xdialog.justify = JUSTIFY_RIGHT;
				break;
			case 48: /* --center option */
				Xdialog.justify = JUSTIFY_CENTER; 
				break;
			case 49: /* --wrap option */
				Xdialog.wrap = TRUE;
				break;
			case 50: /* --no-wrap option */
				Xdialog.wrap = FALSE;
				break;
			case 51: /* --stdout option */
				Xdialog.output = stdout;
				break;
			case 52: /* --stderr option */
				Xdialog.output = stderr;
				break;
			case 53: /* --no-close option */
				Xdialog.no_close = TRUE;
				break;
			case 54: /* --allow-close option */
				Xdialog.no_close = FALSE;
				break;
			case 55: /* --buttons-style option */
				if (!strcmp(optarg, "default"))
					Xdialog.buttons_style = ICON_AND_TEXT;
				else if (!strcmp(optarg, "icon"))
					Xdialog.buttons_style = ICON_ONLY;
				else if (!strcmp(optarg, "text"))
					Xdialog.buttons_style = TEXT_ONLY;
				else
					print_help_info(argv[0],
							"bad button style name");
				break;
			case 56: /* --rc-file option */
				strcpysafe(Xdialog.rc_file, optarg,
					   MAX_FILENAME_LENGTH);
				break;
			case 57: /* --separator option */
				Xdialog.separator[0] = optarg[0];
				break;
			case 58: /* --separate-output option */
				Xdialog.separator[0] = '\n';
				break;
		/* Transient options */
			case 80: /* --fixed-font option */
				Xdialog.fixed_font = TRUE;
				break;
			case 81: /* --passwd option */
				Xdialog.passwd++;
				break;
			case 82: /* --no-cancel option */
				Xdialog.cancel_button = FALSE;
				break;
			case 83: /* --no-button option */
				Xdialog.buttons = FALSE;
				break;
			case 84: /* --no-tags option */
				Xdialog.tags = FALSE;
				break;
			case 85: /* --help option */
				strcpysafe(help_text, optarg,
					   MAX_LABEL_LENGTH);
				Xdialog.help = TRUE;
				break;
			case 86: /* --print option */
				strcpysafe(Xdialog.printer, optarg,
					   MAX_PRTNAME_LENGTH);
				Xdialog.print = TRUE;
				break;
			case 87: /* --wizard option */
				Xdialog.wizard = TRUE;
				break;
			case 88: /* --default-no option */
				Xdialog.default_no = TRUE;
				break;
			case 89: /* --icon option */
				strcpysafe(Xdialog.icon_file, optarg,
					   MAX_FILENAME_LENGTH);
				Xdialog.icon = TRUE;
				break;
			case 90: /* --interval option */
				Xdialog.interval = -1;
				if (optarg != NULL)
					Xdialog.interval = atoi(optarg);
				if (Xdialog.interval < 0)
					print_help_info(argv[0],
							"bad interval number");
				break;
		/* Special options */
			case 98: /* --version option */
				fprintf(stderr, VERSION"\n");
				return 0;
			case 99: /* --print-version (cdialog compatible) option */
				fprintf(stderr, "Version: "VERSION"\n");
				return 0;
		}
		if (win) {
			gtk_widget_show(Xdialog.window);
			gtk_main();

			switch (Xdialog.exit_code) {
				case 0:
					break;
				case 2:
					if (strlen(help_text) == 0)
						return Xdialog.exit_code;
					Xdialog.help = FALSE;
					strcpy(title_tmp, Xdialog.title);
					strcpy(backt_tmp, Xdialog.backtitle);
					strcpy(Xdialog.title, HELP);
					Xdialog.backtitle[0] = 0;
					icon_tmp = Xdialog.icon;
					Xdialog.icon = FALSE;
					create_msgbox(help_text, FALSE);
					Xdialog.help = TRUE;
					Xdialog.icon = icon_tmp;
					strcpy(Xdialog.title, title_tmp);
					strcpy(Xdialog.backtitle, backt_tmp);
					gtk_widget_show(Xdialog.window);
					gtk_main();
					optind = old_optind;
					if (Xdialog.exit_code == 0)
						goto show_again;
				default:
					return Xdialog.exit_code;
			}

			/* Reset transient options flags to their default values */
			Xdialog.passwd		= 0;
			Xdialog.interval	= 0;
			Xdialog.icon		= FALSE;
			Xdialog.buttons		= TRUE;
			Xdialog.cancel_button	= TRUE;
			Xdialog.help		= FALSE;
			Xdialog.print		= FALSE;
			Xdialog.default_no	= FALSE;
			Xdialog.wizard		= FALSE;
			Xdialog.tags		= TRUE;
			Xdialog.fixed_font	= FALSE;
		}
	}

	if (!win)
		print_help_info(argv[0], "missing box option");
	return 0;
}
